'''
BoatViz User Interface for The Bandit Naval warfare RTS game
'''

import wx
import math

from OpenGL.GLUT import *
from OpenGL.GLU import *
from OpenGL.GL import *

class Texture2D:

    nextTextureID = 0
    loadedTextures = {} 

    def __init__(self, filename):

        self.trace = False
        #Check if the image has already been loaded
        if filename in Texture2D.loadedTextures:
            self.textureID, self.img = Texture2D.loadedTextures[filename]

        else: 
            self.loadwx(filename)
            Texture2D.loadedTextures[filename] = (self.textureID, self.img)


    #Load the texture using a wxPython image
    def loadwx(self, filename):
        if self.trace: print "Loading image file %s" % filename

        glEnable(GL_TEXTURE_2D)

        #Create a unique texid
        self.textureID = self.__class__.nextTextureID
        self.__class__.nextTextureID += 1

        #Load terrain texture
        glBindTexture(GL_TEXTURE_2D, self.textureID)

        if filename[-3:] == 'jpg':

            #Load the image file to a wx object
            self.img = wx.Image(filename, wx.BITMAP_TYPE_JPEG)

            #wxPython messes with th images if they are not factors of 4 in width
            #This is probably becuase its built around bitmaps which require padding 
            #so the internal representation is padded, which openGL doesn't like. 
            if (self.img.GetWidth() % 4 != 0) :
                print "The image %s is an unsupported width (%d x %d pixels). Please resize it to a factor of four in width (or strip off %d characters)" % (filename, self.img.GetWidth(), self.img.GetHeight(), self.img.GetWidth()%4) 
                exit()
    
    
            #Configure and blit the texture data to the graphics card
            glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR) 
            glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR) 
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, self.img.GetWidth(), self.img.GetHeight(),
                0, GL_RGB, GL_UNSIGNED_BYTE, self.img.GetData())
        elif filename[-3:] == 'png':
            
            #Load the PNG file to a wx obj
            self.img = wx.Image(filename, wx.BITMAP_TYPE_PNG)

            #wxPython messes with th images if they are not factors of 4 in width
            #This is probably becuase its built around bitmaps which require padding 
            #so the internal representation is padded, which openGL doesn't like. 
            if (self.img.GetWidth() % 4 != 0) :
                print "The image %s is an unsupported width (%d x %d pixels). Please resize it to a factor of four in width (or strip off %d characters)" % (filename, self.img.GetWidth(), self.img.GetHeight(), self.img.GetWidth()%4) 
                exit()

            if not self.img.HasAlpha():
                print "The image %s does not have an alpha channel - use JPG for non-alpha'd images" % filename
                exit()

            #We have to manually build the RGBA data file becuase wx.Image is stupid and wont just
            #give it to us!
            rgba = ""
            count = 0
            imgdata = self.img.GetData()
            for a in self.img.GetAlphaData():
                rgba += imgdata[count:count+3]
                rgba += a
                count += 3

            #Configure and blit the texture data to the graphics card
            glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR) 
            glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR) 

            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, self.img.GetWidth(), self.img.GetHeight(),
                0, GL_RGBA, GL_UNSIGNED_BYTE, rgba)


        else:
            print "The filename extension %s on %s is not supported sorry" % (filename[-3:], filename)
            exit()

    
        glDisable(GL_TEXTURE_2D)

    def getPixelWidth(self):
        return self.img.GetWidth()

    def getPixelHeight(self):
        return self.img.GetHeight()



    def draw(self, width=1, height=1, x=0, y=0):

        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, self.textureID)
        glColor3f(1, 1, 1)

        glPushMatrix()
        glTranslatef(x, y, 0)


        glBegin(GL_QUADS)

        glTexCoord2f(0.0, 1.0)
        glVertex3f(0.0, 0.0, 0.0)

        glTexCoord2f(1.0, 1.0)
        glVertex3f(width, 0.0, 0.0)

        glTexCoord2f(1.0, 0.0)
        glVertex3f(width, height, 0.0)
        
        glTexCoord2f(0.0, 0.0)
        glVertex3f(0.0, height, 0.0)
        glEnd()

        glPopMatrix()

        glDisable(GL_TEXTURE_2D)



circleLookup = None


def drawCircle(x, y, radius=1, lineWidth=1, color=(1, 1, 1)):

    global circleLookup

    if not circleLookup:
        circleLookup = []
        for i in range(0, 360, 30):
            rad = i*math.pi/180.0
            circleLookup.append((math.cos(rad), math.sin(rad)))

    glPushMatrix()

    r, g, b = color
    glColor3f(r, g, b)

    glLineWidth(lineWidth)

    glTranslatef(x, y, 0)
    glBegin(GL_LINE_LOOP)

    for i in circleLookup:
        glVertex3f(i[0]*radius, i[1]*radius, 0)

    glEnd()
    glColor3f(1, 1, 1)

    glPopMatrix()


def drawSelectionRing(x, y, radius=1, color=(1, 1, 1), filled=True, rotation=0):
    '''
    Draws a pretty ring (around an object) to help to visually highlight it. Rather
    than a plain circle, I draw a nicer shape with an outter circle and inner ring 
    made up of segments
    '''

    glPushMatrix()

    r, g, b = color
    glColor3f(r, g, b)

    glLineWidth(1)

    glTranslatef(x, y, 0)

    #Rotation it so allow animation of the selection rings
    glRotatef(rotation, 0, 0, 1)

    if filled:
        #Filled ring is drawn with a triangle strip wih a single pass
        glBegin(GL_TRIANGLE_STRIP)

        for i in range(0, 363, 3):
            rad = i*math.pi/180.0
            #Outer circle drawn 40% larger than the supplied radius
            glVertex3f(math.cos(rad)*radius*1.4, math.sin(rad)*radius*1.4, 0)

            #The inner segment is broken up into 16 sections (alternates every 360/16 degrees)
            #These segments alternate between radius and radius+20%
            if (i/(360/16) % 2) == 0:
                glVertex3f(math.cos(rad)*radius*1.2, math.sin(rad)*radius*1.2, 0)
            else:   
                glVertex3f(math.cos(rad)*radius*1.0, math.sin(rad)*radius*1.0, 0)

        glEnd()

    else: 

        #The non-filled graphic is drawn with the outter and inner done separately

        #Draw the outter plain circle
        glBegin(GL_LINE_LOOP)
        for i in range(0, 360, 3):
            rad = i*math.pi/180.0
            glVertex3f(math.cos(rad)*radius*1.4, math.sin(rad)*radius*1.4, 0)
        glEnd()

        #Draw the inner shape
        glBegin(GL_LINE_LOOP)
        for i in range(0, 360, 3):
            rad = i*math.pi/180.0
            if (i/(360/16) % 2) == 0:
                glVertex3f(math.cos(rad)*radius*1.2, math.sin(rad)*radius*1.2, 0)
            else:   
                glVertex3f(math.cos(rad)*radius*1.0, math.sin(rad)*radius*1.0, 0)

        glEnd()

    glColor3f(1, 1, 1)

    glPopMatrix()

def drawRectangle(x, y, width, height, color, filled=False, alpha=1.0):
    
    r, g, b = color
    glColor4f(r, g, b, alpha)

    glLineWidth(1)

    glPushMatrix()
    glTranslatef(x, y, 0)

    if filled:
        glBegin(GL_QUADS)
    else:
        glBegin(GL_LINE_LOOP)
    glVertex3f(0, 0, 0)
    glVertex3f(width, 0, 0)
    glVertex3f(width, height, 0)
    glVertex3f(0, height, 0)
    glEnd()

    glPopMatrix()

def drawLine(x1, y1, x2, y2, color):

    r, g, b = color
    glColor3f(r, g, b)

    glLineWidth(1)

    glBegin(GL_LINES)
    glVertex3f(x1, y1, 0)
    glVertex3f(x2, y2, 0)
    glEnd()

    






