import random

ENEMY_NUM = 5
FRIENDLY_NUM = 5
ENEMY_RANGE = 8
# value for each entity (enemy and friendly utils, and the flag), 
# times by the number of maps to create
GENENUM = (ENEMY_NUM + FRIENDLY_NUM +1) * FRIENDLY_NUM 

class Individual:

    # constructor
    def __init__(self):
        global ENEMY_NUM, FRIENDLY_NUM
        self.chromosome = [] 
        self.fitness = 0.0
        self.wins = 0
        self.losses = 0
        self.gamesplayed = 0
        self.fitnessAttributes = []

        self.attacker = True

        self.trace = False

    # comparable method
    def __cmp__(self, other):
        return(cmp(self.getFitness(), other.getFitness()))

    def initialize(self):
    # define a random number for the given genes
    # the genes represent a value for each useful 'entity' in the 
    # IM (e.g. enemy units, friendly units) 

        global GENENUM
        for i in range(GENENUM):
            geneVal = random.randint(-50,50)
            self.chromosome.append(geneVal)

    # used for debugging, set a specific chromosome
    def setChromosome(self):

        # attacker chromosome
        if self.attacker:
            self.chromosome = [ 
            35, 26, 40, -31, 27, -25, 32, -9, 31, 45, 31, 
            26, 45, 20, 25, 34, 14, -16, 6, 17, 23, -28, 
            46, -13, 20, -22, -10, 46, 9, -22, -41, 47, -15, 
            13, -33, -7, -21, -34, 4, -12, -43, -26, -13, -30, 
            -13, -21, 48, 18, 33, 49, 23, -15, -44, -38, -46 ]
        else:
        # defender chromosome
            self.chromosome = [ 
                4, -32, 22, 17, 8, -37, 19, 45, -35, -30, -40, 
                -21, -47, 28, 50, 27, 47, 31, -2, 36, 6, 30, 
                44, -4, -42, -6, -47, -26, -48, 18, 10, 26, 16, 
                -12, -26, 12, 47, -37, 43, -20, -3, -8, 7, -2, 
                -9, 32, -50, 2, -41, 16, 45, -48, -32, -39, -39 ]
			
        
    def generateMaps(self, map):
        ''' use the passed map matrix to generate a map '''

        global FRIENDLY_NUM, ENEMY_NUM, ENEMY_RANGE

        # for coevolution, an individual can be an attacker or defender.
        # if it's an attacker, the ranged cell values should be > FRIENDLY_NUM for the defenders
        friendRangeLwr = 0
        friendRangeHigh = 0
        enemyRangeLwr = 0
        enemyRangeHigh = 0
        if self.attacker:
            friendRangeLwr = 1
            friendRangeHigh = FRIENDLY_NUM
            enemyRangeLwr = FRIENDLY_NUM + 1
            enemyRangeHigh = FRIENDLY_NUM + ENEMY_NUM
        else:
            friendRangeLwr = FRIENDLY_NUM + 1
            friendRangeHigh = FRIENDLY_NUM + ENEMY_NUM
            enemyRangeLwr = 1
            enemyRangeHigh = FRIENDLY_NUM

        toreturn = [] 
        for mapNo in range(FRIENDLY_NUM):
            returnMap = []
            for i in range(len(map)):
                returnMap.append([])
                for j in range(len(map[i])):
                    returnMap[i].append(0)


            for col in range(len(map)):
                for row in range(len(map[col])):
                # check if the cell is occupied by a enemy or unit
                    cell = map[col][row]
                    if (cell >= friendRangeLwr and cell <= friendRangeHigh) or cell == -1:
                    # friendly unit or flag unit
                        if cell == -1:
                            value = self.chromosome[(ENEMY_NUM + FRIENDLY_NUM) + ((ENEMY_NUM + FRIENDLY_NUM +1) * mapNo)];
                        else:
                            value = self.chromosome[(cell-1) + ((ENEMY_NUM + FRIENDLY_NUM +1) * mapNo)];
                        returnMap[col][row] += value;
                    elif cell >= enemyRangeLwr and cell <= enemyRangeHigh:
                        #enemy unit
                        value = self.chromosome[(cell-1) + ((ENEMY_NUM + FRIENDLY_NUM +1) * mapNo)];
                        returnMap[col][row] += value;
                        #permiate the range of the enemy value
                        for incol in range(1, ENEMY_RANGE):
                            for inrow in range(1, ENEMY_RANGE):
                                newValue = value - 2 * (value/ENEMY_RANGE * 
                                (max(inrow,incol)))
                                if value > 0 and value <= 50:
                                    if newValue <= 0:
                                        newValue= 1
                                elif value < 0 and value >= -50:
                                    if newValue >= 0:
                                        newValue = -1
                                            
                                if col-incol >=0:
                                    if row-inrow >= 0:
                                        returnMap[col-incol][row-inrow] += newValue
                                    if row+inrow < len(map[col-incol]):
                                        returnMap[col-incol][row+inrow] += newValue
                                if col+incol < len(map):
                                    if row+inrow < len(map[col+incol]):
                                        returnMap[col+incol][row+inrow] += newValue
                                    if row-inrow >= 0:
                                        returnMap[col+incol][row-inrow] += newValue

                            newValue = value - 2 * (value/ENEMY_RANGE * incol)
                            if value > 0 and value <= 50:
                                if newValue <= 0:
                                    newValue= 1
                            elif value < 0 and value >= -50:
                                if newValue >= 0:
                                    newValue= -1


                            if(row-incol >= 0):
                                returnMap[col][row-incol] += newValue
                            if(row+incol < len(map[col])):
                                returnMap[col][row+incol] += newValue
                            if(col-incol >= 0):
                                returnMap[col-incol][row] += newValue
                            if(col+incol < len(map)):
                                returnMap[col+incol][row] += newValue 

            toreturn.append(returnMap)
        
        return toreturn

    def setGene(self, index, value):
        " update a value of the chromosome "
        self.chromosome[index] = value

    
    def getGene(self, index):
        " retrieve value of the chromosome "
        return self.chromosome[index]

    def getWinRatio(self):
        if self.gamesplayed > 0:
            return float(self.wins) / float(self.gamesplayed)
        return 0

    def setFitness(self, win, playerCount, enemyCount, time, opponentWinRatio, coevolve):
        if self.trace: print "Fitness:  %s, %i, %i, %f" % (str(win), playerCount, enemyCount, time)

        self.gamesplayed+=1
        
        if win:
            self.wins+=1
        else:
            self.losses+=1

        if coevolve:
            winratio = float(float(self.wins) / float(self.gamesplayed)) * 500
            if self.trace: print "coevolved winratio: %f wins: %i gamesplayed: %i" % (winratio, self.wins, self.gamesplayed)
        else:
            if win:
                winratio = 500
            else:
                winratio = 0

        self.fitness = winratio + 5 * playerCount - 20 * enemyCount + time
        if self.trace: print "PlayerCount: %f EnemyCount: %f Time: %f " % ((5 * playerCount), (20 * enemyCount), time)
        
        if coevolve:
            if opponentWinRatio != 0:
                # the player should be rewarded for playing a harder opponent with a
                # better win ratio
                self.fitness += opponentWinRatio * 100
                if self.trace: print "OpponentRatio: " + str(opponentWinRatio * 100)

        # bonus for attackers - only when not coevolving
        if not coevolve and enemyCount == 0:
            self.fitness += 50

        self.fitnessAttributes = str(win) + "\t" + str(playerCount) + "\t" + str(enemyCount) + "\t" + str(time)


    def getFitness(self):
        return self.fitness

    def fitnessToString(self):
        return self.fitnessAttributes;

    def toString(self):
        retStr = "[ "
        for i in self.chromosome:
            retStr += str(i)
            retStr += ", "
        retStr += "]"
        return retStr
            

'''
# use the following to test the working of the class - 
# just type python individual.py on command line
# used to test generation of genes and maps
if __name__=="__main__":
    random.seed()
    global GENENUM, ENEMY_NUM, FRIENDLY_NUM
    ind = Individual()
    ind.initialize()

    print("Individual:")
    line = ""
    for i in range(GENENUM):
        line += str(ind.getGene(i))
        line += ",  "
        if (i+1) % (ENEMY_NUM + FRIENDLY_NUM +1) == 0:
            line += "| "
    print line

    #map = [ [1, 0,0,0,0,0,0,0], [2,0,0,0,0,0,0,0], [3,0,0,0,8,0,0,0], [4,0,0,0,0,0,0,0], [5,0,0,0,0,0,0,0] ]
    map = [ [1, 0,0,0,0,6,0,0], [2,0,0,0,7,0,0,0], [3,0,0,0,8,0,0,0], [4,0,0,0,9,0,0,0], [5,0,0,0,0,10,0,0] ]
    print("Before")
    line = ""
    for i in range(len(map)):
        for j in range(len(map[i])):
            line += str(map[i][j])
            line += "\t"
        print ("[ " + line + " ]")
        line = ""

    returned = ind.generateMaps(map)

    line = ""
    print("After: ")
    for maps in returned:
        for i in range(len(maps)):
            for j in range(len(maps[i])):
                line += str(maps[i][j])
                line += "\t"
            print ("[ " + line + " ]")
            line = ""
        print("")
        line = ""
'''
