import graphics
import vector

import random

import wx
from OpenGL.GL import *

class MapCell:
    def __init__(self, x, y):
        self.parent = None
        self.cost = None
        self.estCost = None
        self.weight = 0

        self.closed = False
        self.opened = False

        self.x = x
        self.y = y

        self.visitable = True

    def cmp(a, b):
        if a.estCost == None: aa = 0
        else: aa = a.estCost

        if b.estCost == None: bb = 0    
        else: bb = b.estCost

        if aa<bb: return -1
        if aa==bb: return 0
        if aa>bb: return 1


class InfluenceMap:

    def __init__(self, realWidth, realHeight, gridWidth, gridHeight): 

        self.realWidth = realWidth
        self.realHeight = realHeight
        self.gridWidth = gridWidth
        self.gridHeight = gridHeight

        self.map = []
        for x in range(gridWidth):
            self.map.append([])
            for y in range(gridHeight):
                self.map[x].append(MapCell(x, y))

        '''
        terrainMapImg = globals.scenario.terrainInfluence
        for y in range(terrainMapImg.GetHeight()):
            for x in range(terrainMapImg.GetWidth()):
                r = terrainMapImg.GetRed(x, y)
                if r == 0:
                    self.map[x][self.gridHeight-y-1].visitable = True
        '''


    def setWeights(self, weightmap):

        for x in range(self.gridWidth):
            for y in range(self.gridHeight):
                self.map[x][y].weight = weightmap[x][y]




    def findLowestWeight(self, scenario):
        pos = (0, 0)
        weight = self.map[0][0].weight

        for i in self.map:
            for j in i:
                if j.weight < weight:
                    weight = j.weight
                    pos = (self.map.index(i), i.index(j))

        if weight == self.map[0][0].weight:
           return (scenario.flagx, scenario.flagy) 
        return self.gridToWorld(pos[0], pos[1])
    

    def findPath(self, x1, y1, x2, y2):

        self.resetMap()

        destx, desty = self.worldToGrid(x2, y2)
        dest = self.map[destx][desty]

        startx, starty = self.worldToGrid(x1, y1)

        #If the boat is out of bounds for some reason - make sure we are only routing from a valid
        #location
        if startx<0: startx = 0
        if startx>=self.gridWidth: startx = self.gridWidth-1
        if starty<0: starty = 0
        if starty>=self.gridHeight: starty = self.gridHeight-1
            
        start = self.map[startx][starty]

        start.cost = 0
        start.estCost = max(abs(dest.x-start.x), abs(dest.y-start.y))


        open = [start]

        while not dest in open:
            open.sort(MapCell.cmp)

            curr = open[0]

            newcells = self.getValidSurroundingCells(curr.x, curr.y)
            for i in newcells:
                i.cost = curr.cost+vector.Vector(i.x-curr.x, i.y-curr.y).length()+i.weight
                i.estCost = i.cost + vector.Vector(dest.x-i.x, dest.y-i.y).length()
                i.parent = curr
                i.opened = True
                open.append(i)

            curr.closed = True
            open.remove(curr)


        curr = dest
        bestPath = []
        while not curr == start:
            bestPath.insert(0, self.gridToWorld(curr.x, curr.y))

            curr = curr.parent


        return bestPath

    def resetMap(self):
        for i in self.map:
            for j in i:
                j.cost = None
                j.estCost = None 
                j.parent = None
                j.closed = False
                j.opened = False


    def getValidSurroundingCells(self, x, y):
        cells = []
        c = self.getCell(x-1, y-1); 
        if c and c.visitable and not c.closed and not c.opened: cells.append(c)
        c = self.getCell(x-1, y  ); 
        if c and c.visitable and not c.closed and not c.opened: cells.append(c)
        c = self.getCell(x-1, y+1); 
        if c and c.visitable and not c.closed and not c.opened: cells.append(c)
        c = self.getCell(x  , y-1); 
        if c and c.visitable and not c.closed and not c.opened: cells.append(c)
        c = self.getCell(x  , y+1); 
        if c and c.visitable and not c.closed and not c.opened: cells.append(c)
        c = self.getCell(x+1, y-1); 
        if c and c.visitable and not c.closed and not c.opened: cells.append(c)
        c = self.getCell(x+1, y  ); 
        if c and c.visitable and not c.closed and not c.opened: cells.append(c)
        c = self.getCell(x+1, y+1); 
        if c and c.visitable and not c.closed and not c.opened: cells.append(c)
        return cells

        
    def getCell(self, x, y):
        if 0 <= x < self.gridWidth and 0 <= y < self.gridHeight:
            return self.map[x][y]
        else:
            return None


        

    def drawDebug(self):
        xSpacing = float(self.realWidth) / self.gridWidth
        ySpacing = float(self.realHeight) / self.gridHeight

        glBegin(GL_LINES)
        glColor4f(1, 1, 1, 0.2)
        #draw cols
        for i in range(self.gridWidth):
            glVertex2f(i*xSpacing, 0)
            glVertex2f(i*xSpacing, self.realHeight)
        #draw rows
        for i in range(self.gridHeight):
            glVertex2f(0, i*ySpacing)
            glVertex2f(self.realWidth, i*ySpacing)
        glEnd()

        glColor4f(1, 1, 1, 1)

        for y in range(self.gridHeight):
            for x in range(self.gridWidth):
                wx, wy = self.gridToWorld(x, y)
                if not self.map[x][y].visitable:
                    graphics.drawLine(wx-50, wy-50, wx+50, wy+50, color=(1, 0, 0))
                    graphics.drawLine(wx+50, wy-50, wx-50, wy+50, color=(1, 0, 0))

                if self.map[x][y].weight > 0:
                    graphics.drawRectangle(wx-50, wy-50, 100, 100, color=(0, 1, 0), filled=True, alpha=self.map[x][y].weight*2.0/100.0) 
                elif self.map[x][y].weight < 0:
                    graphics.drawRectangle(wx-50, wy-50, 100, 100, color=(1, 0, 0), filled=True, alpha= -self.map[x][y].weight*2.0/100.0) 

    def gridToWorld(self, x, y):
        xSpacing = float(self.realWidth) / self.gridWidth
        ySpacing = float(self.realHeight) / self.gridHeight

        worldx = x * (float(self.realWidth)/self.gridWidth) + xSpacing/2
        worldy = y * (float(self.realHeight)/self.gridHeight) + ySpacing/2
        return (worldx, worldy)

    def worldToGrid(self, x, y):
        gridx = x * float(self.gridWidth)/self.realWidth
        gridy = y * float(self.gridHeight)/self.realHeight

        if gridx<0: gridx = 0
        if gridx>=self.gridWidth: gridx = self.gridWidth-1
        if gridy<0: gridy = 0
        if gridy>=self.gridHeight: gridy = self.gridHeight-1

        return(int(gridx), int(gridy))

        
