from mpi4py import MPI
import numpy
import population
import individual
import sys
import globals
import scenario
import entity
import random


START=0
DATA=1
FINISHED=666

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

#master process
print "Rank: " + str(rank)
print "Size: " + str(size)

done = False
chromosomeSize = individual.GENENUM
fitnessCount = 5
ATTACKER = 0
DEFENDER = 1

if rank == 0:
    popIndex = 0
    generation = 0
    runNum = 0


    # the population 
    pop = population.Population("")

    # for coevolution
    pop1 = population.Population("Attacker")
    pop2 = population.Population("Defender")
    pop2.setDefenders()
    evolve1 = True

    #pop1.setChromosomes()
    #pop2.setChromosomes()


    # running list
    nodesRun = []

    # run a certain number of runs
    while runNum < 30:
        #generation loop
        while generation < population.GENERATIONS:
            print "Generation: " + str(generation)
            #evaluate all individuals in the population using cluster
            # use all nodes bar the root
            startdata = numpy.arange(2, dtype=int)
            finishdata = numpy.arange(fitnessCount, dtype=float)
            # check each node to see if a message has been sent to start an evaluation
            # use non blocking recieve, and check if it has been received.

            # --------------------------------------
            # Coevolutionary generation
            # --------------------------------------
            if globals.COEVOLVE:
                while popIndex < population.POP_SIZE:
                    # check for starting node
                    for i in range(1,size):
                        if comm.Iprobe(source=i, tag=START):
                            print "Start received"
                            # node ready to evaluate
                            comm.Recv([startdata, MPI.INT], source=i, tag=START)
                            randIndex = random.randint(0, population.POP_SIZE - 1)
                            print "RandIndex: " + str(randIndex)
                            popType = 0
                            if evolve1:
                                # Population 1 eval - Attacker population
                                indiv1 = pop1.getIndividual(popIndex)
                                indiv2 = pop2.getIndividual(randIndex)
                                popType = ATTACKER
                            else:
                                # Population 2 eval - Defender population
                                indiv1 = pop1.getIndividual(randIndex)
                                indiv2 = pop2.getIndividual(popIndex)
                                popType = DEFENDER
                            popIndex+=1
                            # send data is of size chromosomeSize * 2 due to 2 individuals, +1 for the 
                            # popindex at the end + 2 for poptype and oppindex
                            sendData = numpy.arange(chromosomeSize*2 + 3, dtype=int)
                            for gene in range(chromosomeSize):
                                sendData[gene] = indiv1.getGene(gene)
                                sendData[gene+chromosomeSize] = indiv2.getGene(gene)
                            sendData[-1] = (popIndex-1)
                            sendData[-2] = popType
                            sendData[-3] = randIndex

                            comm.Send([sendData, MPI.FLOAT], dest=i, tag=DATA)
                            nodesRun.append(i)
                    # check for finished node
                    for i in range(1, size):
                        if comm.Iprobe(source=i, tag=FINISHED):
                        # evaluation finished, returning fitness results
                            finishdata = numpy.arange(fitnessCount + 2, dtype=float)
                            comm.Recv([finishdata, MPI.FLOAT], source=i, tag=FINISHED)
                            popType = int(finishdata[-2])
                            oppIndex = int(finishdata[-1])
                            if popType == ATTACKER:
                                pop1.setFitness(int(finishdata[0]), int(finishdata[1]), int(finishdata[2]), int(finishdata[3]), float(finishdata[4]), pop2.getIndividual(oppIndex).getWinRatio())
                                print "Setting fitness with: %i %s %i %i %f %f" % ( int(finishdata[0]), int(finishdata[1]), int(finishdata[2]), int(finishdata[3]), float(finishdata[4]), pop2.getIndividual(oppIndex).getWinRatio())
                            elif popType == DEFENDER:
                                pop2.setFitness(int(finishdata[0]), not int(finishdata[1]), int(finishdata[3]), int(finishdata[2]), float(finishdata[4]), pop1.getIndividual(oppIndex).getWinRatio())
                            else:
                                print "ERROR: poptype not set, or opponent index not set"
                                print "Node index: %i Poptype: %s oppIndex %s" % (i, str(popType), str(oppIndex))
                                sys.exit(1)
                            # pop off running list
                            if nodesRun.count(i) > 0:
                                nodesRun.remove(i)

                # wait for population to finish evaluation
                print "waiting"
                for i in nodesRun:
                    comm.Recv([finishdata, MPI.FLOAT], source=i, tag=FINISHED)
                    
                    popType = int(finishdata[-2])
                    oppIndex = int(finishdata[-1])

                    if popType == ATTACKER:
                        pop1.setFitness(int(finishdata[0]), int(finishdata[1]), int(finishdata[2]), int(finishdata[3]), float(finishdata[4]), pop2.getIndividual(oppIndex).getWinRatio())
                        print "Setting fitness with: %i %s %i %i %f %f" % ( int(finishdata[0]), int(finishdata[1]), int(finishdata[2]), int(finishdata[3]), float(finishdata[4]), pop2.getIndividual(oppIndex).getWinRatio())
                    elif popType == DEFENDER:
                        pop2.setFitness(int(finishdata[0]), not int(finishdata[1]), int(finishdata[3]), int(finishdata[2]), float(finishdata[4]), pop1.getIndividual(oppIndex).getWinRatio())
                    else:
                        print "ERROR: poptype not set, or opponent index not set"
                        print "Node index: %i Poptype: %s oppIndex %s" % (i, str(popType[i]), str(oppIndex))
                        sys.exit(1)
            
                print "done waiting"

                # finished the population loop
                # all individuals evaluated, start next generation
                popIndex = 0
                if evolve1:
                    pop1.newPopulation(generation)
                    evolve1 = False
                else:
                    pop2.newPopulation(generation)
                    evolve1 = True
                    generation += 1
                nodesRun = []

            # --------------------------------------
            # No coevolution generation
            # --------------------------------------
            else:
                while popIndex < population.POP_SIZE:
                    for i in range(1,size):
                        if comm.Iprobe(source=i, tag=START):
                        # request for evaluation made
                            comm.Recv([startdata, MPI.INT], source=i, tag=START)
                            indiv = pop.getIndividual(popIndex)
                            popIndex+=1
                            sendData = numpy.arange(chromosomeSize +1, dtype=float)
                            for gene in range(chromosomeSize):
                                sendData[gene] = indiv.getGene(gene)
                            sendData[-1] = (popIndex-1)
                            comm.Send([sendData, MPI.FLOAT], dest=i, tag=DATA)
                            nodesRun.append(i)
                    for i in range(1, size):
                        if comm.Iprobe(source=i, tag=FINISHED):
                        # evaluation finished, returning fitness results
                            comm.Recv([finishdata, MPI.FLOAT], source=i, tag=FINISHED)
                            pop.setFitness(int(finishdata[0]), int(finishdata[1]), int(finishdata[2]), int(finishdata[3]), float(finishdata[4]))
                            # pop off running list
                            if nodesRun.count(i) > 0:
                                nodesRun.remove(i)

                # wait for population to finish evaluation
                print "waiting"
                for i in nodesRun:
                    comm.Recv([finishdata, MPI.FLOAT], source=i, tag=FINISHED)
                    pop.setFitness(int(finishdata[0]), int(finishdata[1]), int(finishdata[2]), int(finishdata[3]), float(finishdata[4]))
                print "done waiting"

                # all individuals evaluated, start next generation
                pop.newPopulation()
                popIndex = 0
                generation +=1
                nodesRun=[]

        popIndex = 0
        generation = 0

        # the population 
        pop = population.Population("")

        # for coevolution
        pop1 = population.Population("Attacker")
        pop2 = population.Population("Defender")
        pop2.setDefenders()
        evolve1 = True

        #pop1.setChromosomes()
        #pop2.setChromosomes()


        # running list
        nodesRun = []

        runNum += 1
        print "Run number: " + str(runNum)

# slave process
else:
    while True:

        # send for data
        data = numpy.arange(2, dtype=int)
        data[0] = rank
        print "Ready to start for : " + str(rank)
        comm.Send([data, MPI.INT], dest=0, tag=START)
        if globals.COEVOLVE:
            chromosome = numpy.arange(chromosomeSize*2 + 3, dtype=int)
        else:
            chromosome = numpy.arange(chromosomeSize + 1, dtype=int)
        comm.Recv([chromosome, MPI.FLOAT], source=0, tag=DATA)
        popindex = chromosome[-1]
        popType = chromosome[-2]
        oppIndex = chromosome[-3]
        print "PopIndex: " + str(popindex)

        string = ""
        for i in chromosome:
            string += str(i) + ", "
        print "Received: " + string

        # deltatime stuff
        dt = 0.1

        # reset global stuff
        globals.reset()

        if globals.COEVOLVE:
            indiv1 = individual.Individual()
            indiv1.initialize()
            indiv2 = individual.Individual()
            indiv2.initialize()
            indiv2.attacker = False

            string1 = ""
            string2 = ""
            for i in range(chromosomeSize):
                indiv1.setGene(i, chromosome[i])
                string1 += str(chromosome[i]) + ", "
                indiv2.setGene(i, chromosome[i + chromosomeSize])
                string2 += str(chromosome[i + chromosomeSize]) + ", "

            print "Individual1: " + string1
            print "Individual2: " + string2

            globals.scenario = scenario.Scenario('data/island_scenario.dat')
            globals.scenario.noGraphInitialize()

            globals.entityManager.attacker = indiv1
            globals.entityManager.defender = indiv2

            print "starting game"

            while True:
                globals.step()
                globals.gameEngine.step(globals.entityManager)
                globals.entityManager.update()
                globals.entityManager.moveEntities()

                # if the game has ended
                if globals.entityManager.gameOver():
                    print "Game finished"
                    playerCount = enemyCount = 0
                    for i in globals.entityManager.entities:
                        if i.teamID == 1:
                            if i.alive:
                                playerCount += 1
                        if i.teamID == 2:
                            if i.alive:
                                enemyCount += 1
                    print str(popindex) + " < " + str(population.POP_SIZE)
                    break

                dt = 0.1
            print "Finished evaluation"

        else:
            
            # set up the game objects (order is important)
            indiv = individual.Individual()
            indiv.initialize()

            for i in range(chromosomeSize):
                indiv.setGene(i, chromosome[i])

            globals.scenario = scenario.Scenario('data/island_scenario.dat')
            globals.scenario.noGraphInitialize()

            while True:

                globals.step()
                globals.gameEngine.step(globals.entityManager)
                globals.entityManager.update()
                globals.entityManager.moveEntities()

                # if the game has ended
                if globals.entityManager.gameOver():
                    playerCount = enemyCount = 0
                    for i in globals.entityManager.entities:
                        if i.teamID == 1:
                            if i.alive:
                                playerCount += 1
                        if i.teamID == 2:
                            if i.alive:
                                enemyCount += 1
                    print str(popindex) + " < " + str(population.POP_SIZE)
                    break

                dt = 0.1
       
        sendData = numpy.arange(fitnessCount + 2, dtype=float)
        
        sendData[0] = popindex
        sendData[1] = globals.entityManager.win
        sendData[2] = playerCount
        sendData[3] = enemyCount
        sendData[4] = globals.entityManager.timeLeft()
        sendData[5] = popType
        sendData[6] = oppIndex 

        comm.Send([sendData, MPI.FLOAT], dest=0, tag=FINISHED)

