'''
BoatViz User Interface for The Bandit Naval warfare RTS game
'''


#wxPython imports
import wx
from wx.glcanvas import GLCanvas 

#OpenGL imports
from OpenGL.GLUT import *
from OpenGL.GLU import *
from OpenGL.GL import *

#Python libs
import sys,math
import random

#Local libraries
import graphics
import globals
import population
import individual
import scenario

initialWidth = 1024
initialHeight = 768

class myGLCanvas(GLCanvas):

    def __init__(self, parent):
        GLCanvas.__init__(self, parent,-1, size=(initialWidth,initialHeight))
        wx.EVT_PAINT(self, self.onPaint)
        self.init = 0

        self.Bind(wx.EVT_LEFT_DOWN, self.mouseEvent)
        self.Bind(wx.EVT_LEFT_UP, self.mouseEvent)
        self.Bind(wx.EVT_RIGHT_DOWN, self.mouseEvent)
        self.Bind(wx.EVT_MIDDLE_DOWN, self.mouseEvent)
        self.Bind(wx.EVT_MOTION, self.mouseEvent)
        self.Bind(wx.EVT_MOUSEWHEEL, self.mouseEvent)
        self.Bind(wx.EVT_KEY_DOWN, self.keyPress)
        self.Bind(wx.EVT_IDLE, self.onIdle)
        self.Bind(wx.EVT_SIZE, self.onSize)

        self.entityCursor = None

        self.viewport = Viewport()

        self.selectionBox = None


        return

    def onPaint(self,event):
        dc = wx.PaintDC(self)
        self.SetCurrent()
        if not self.init:
            self.InitGL()
            self.init=1

        self.setViewPoint()
        self.onDraw()
        return

    def onIdle(self, event):
        self.setViewPoint()
        self.onDraw()

        event.RequestMore()

    def onSize(self, event):
        '''Called when the window is resized'''
        width, height = self.GetSize()
        if self.GetContext():
            self.SetCurrent()
            glViewport(0, 0, width, height)

        self.viewport.resize(width, height)

         

    def onDraw(self):

        # step the game
        globals.step()
        globals.gameEngine.step(globals.entityManager)
        globals.entityManager.update()

        # population code, executed upon game completion
        if globals.entityManager.gameOver():
			sys.exit(0)


        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glDisable(GL_DEPTH_TEST)
        glPushMatrix()
        glColor3f(1, 1, 1)

        #Draw the background terrain  - the sanfran map represents 92km x 40km
        #Image is: 3060 x 1356 px  -   A section of 1932 px measured 58187m  ==> 92160 x 40840

        #self.terrainTex.draw(92160, 40840, -92160/2, -40840/2)

            
        #Draw scenario related properties - terrain/flags etc
        globals.scenario.draw()


        globals.entityManager.renderEntities()

        #Draw the selection box - which will occur if we are dragging
        if self.selectionBox:
            start = self.selectionBox[0]
            end = self.selectionBox[1]
            glColor3f(0, 1, 0)
            glBegin(GL_LINE_LOOP)
            glVertex2f(start[0], start[1])
            glVertex2f(start[0], end[1])
            glVertex2f(end[0], end[1])
            glVertex2f(end[0], start[1])
            glEnd()

            graphics.drawCircle(start[0], start[1], 10*self.viewport.zoom, 2, color=(0, 1, 0))
            graphics.drawCircle(end[0], end[1], 10*self.viewport.zoom, 2, color=(0, 1, 0))


        glPopMatrix()
        self.SwapBuffers()
        return


    def setViewPoint(self):

        # set viewing projection
        glEnable(GL_DEPTH_TEST)
        glClearColor(0.0, 0.0, 0.0, 1.0)
        glClearDepth(1.0)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        self.viewport.setViewport()

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        return;

    
    def InitGL(self):
    
        #self.setViewPoint()
        
        #Load textures

        globals.scenario.initialize()


        #We don't want depth testing, we just make sure we draw using the painters-algorithm ourselves
        glDisable(GL_DEPTH_TEST)

        #Enable blending so our textures draw with nice transparency
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        return


    '''Begin entity placement mode as controlled by the UI buttons'''
    def entityAircraftCarrierPlacementA(self, event):
       globals.entityManager.addAircraftCarrier(1)

    def entityDestroyerPlacementA(self, event):
       globals.entityManager.addDestroyer(1)

    def entityCargoshipPlacementA(self, event):
       globals.entityManager.addCargoship(1)

    def entityAircraftCarrierPlacementB(self, event):
       globals.entityManager.addAircraftCarrier(2)

    def entityDestroyerPlacementB(self, event):
       globals.entityManager.addDestroyer(2)

    def entityCargoshipPlacementB(self, event):
       globals.entityManager.addCargoship(2)




    def mouseEvent(self, event):

        #Make sure the opengl window has focus again otherwise key events get lost
        self.SetFocus()

        width, height = self.GetSize()

        #Convert the screen coords into floating point values [0-1] as they are easier to handle
        #elsewhere
        x = event.GetX()*1.0/width
        y = 1 - (event.GetY()*1.0/height)

        wx, wy = self.viewport.convertToWorldCoords(x, y)

        '''
        if event.LeftDown():
            print "%.0f %.0f" % (wx, wy)
            return
        '''    

        #Let the entitymanager handle the rest of the selection problem
        if event.LeftDown() and not event.ShiftDown():
            #This is a normal selection operation, left-click. Also start a selection area (even if
            #it may be discarded on the up event instantly
            globals.entityManager.pointSelect(wx, wy)
            self.selectionBox = [(wx, wy), (wx, wy)]

        elif event.LeftDown() and event.ShiftDown():
            #This is a group selection operation (shift is held down) so new points are to be added
            #to the selection rather than re-selected

            globals.entityManager.pointAltSelect(wx, wy)
            self.selectionBox = [(wx, wy), (wx, wy)]

        elif event.LeftUp() and not event.ShiftDown():
            #This is the end of a area/dragging selection so we invalidate the current box after
            #passing it to the entitymanager to handle the selections
            if self.area(self.selectionBox) < 400:
                globals.entityManager.pointSelect(wx, wy)
            else:
                globals.entityManager.areaSelect(self.selectionBox)

            self.selectionBox = None

        elif event.LeftUp() and event.ShiftDown():
            #This is the end of a area/dragging selection in which objects are added to the group
            #rather than creating a new one
            globals.entityManager.areaAltSelect(self.selectionBox)
            self.selectionBox = None

        elif event.MiddleDown():
            globals.entityManager.pointDebug(wx, wy)

        elif event.RightDown() and not event.ShiftDown():
            globals.entityManager.pointAction(wx, wy)

        elif event.RightDown() and event.ShiftDown():
            globals.entityManager.pointAltAction(wx, wy)

        if event.Dragging():
            if self.selectionBox:
                #Set a new end-point for the selection box, and also update the selections with
                #anything inside of it
                self.selectionBox[1] = (wx, wy)
                globals.entityManager.areaAltSelect(self.selectionBox)

        #Handle mousewheel events, for zooming in and out. We also re-center the 
        #viewport on the mouse location which is a nice usability feature
        if event.GetWheelRotation() > 0:
            self.viewport.x = wx
            self.viewport.y = wy
            self.viewport.zoomIn()
        elif event.GetWheelRotation() < 0:
            self.viewport.zoomOut()



    def area(self, box):
        start = box[0]
        end = box[1]
        return abs(start[0]-end[0]) * abs(start[1]-end[1])



    def keyPress(self, event):
        if event.GetKeyCode() == wx.WXK_LEFT:
            self.viewport.left()
        if event.GetKeyCode() == wx.WXK_RIGHT:
            self.viewport.right()
        if event.GetKeyCode() == wx.WXK_UP:
            self.viewport.up()
        if event.GetKeyCode() == wx.WXK_DOWN:
            self.viewport.down()
        if event.GetKeyCode() == ord('+'):
            self.viewport.zoomIn()
        if event.GetKeyCode() == ord('-'):
            self.viewport.zoomOut()

    
class Viewport:

    #The percentage of the screen width we move over each key navigation
    KEY_MOVEMENT_AMOUNT = 0.25

    def __init__(self):
        self.x = globals.scenario.terrainWidth/2
        self.y = globals.scenario.terrainHeight/2
        self.pixelWidth = initialWidth
        self.pixelHeight = initialHeight

        #Zoom value refers to meters per pixel
        self.zoom = 10.0
        
        self.recalcViewport()

    def recalcViewport(self):
        
        #Calculate the minimum zoom allowed so that we can't see past the edges of the map
        self.minZoom = max(globals.scenario.terrainWidth/self.pixelWidth, 
                            globals.scenario.terrainHeight/self.pixelHeight)

        if self.zoom > self.minZoom:
            self.zoom = self.minZoom

        self.viewWidth = self.pixelWidth * self.zoom
        self.viewHeight = self.pixelHeight * self.zoom

        #Make sure the viewport havent scrolled past the edges
        self.x = max(self.x, self.viewWidth/2) 
        self.x = min(self.x, globals.scenario.terrainWidth - self.viewWidth/2)
        self.y = max(self.y, self.viewHeight/2) 
        self.y = min(self.y, globals.scenario.terrainHeight - self.viewHeight/2)

    def resize(self, newwidth, newheight):
        self.pixelWidth = newwidth
        self.pixelHeight = newheight
        self.recalcViewport()


    def setViewport(self):
        glOrtho(-(self.viewWidth/2)+self.x, 
            (self.viewWidth/2)+self.x, 
            -(self.viewHeight/2)+self.y, 
            (self.viewHeight/2)+self.y, -1, 1)

    def left(self):
        self.x-=(self.viewWidth*self.KEY_MOVEMENT_AMOUNT)
        self.recalcViewport()

    def right(self):
        self.x+=(self.viewWidth*self.KEY_MOVEMENT_AMOUNT)
        self.recalcViewport()

    def up(self):
        self.y+=(self.viewHeight*self.KEY_MOVEMENT_AMOUNT)
        self.recalcViewport()

    def down(self):
        self.y-=(self.viewHeight*self.KEY_MOVEMENT_AMOUNT)
        self.recalcViewport()

    def zoomIn(self):
        self.zoom /= 2
        self.recalcViewport()

    def zoomOut(self):
        self.zoom *= 2
        self.recalcViewport()

    def zoomAt(self, x, y):
        '''Zoom in but ensure that the position specified stays in the same world position '''
        #screenX = 
        pass

    def convertToWorldCoords(self, x, y):
        '''
        Converts screen coords [0-1] to world coordinates
        '''
        #Put them so 0 is centered on the display
        x-=0.5
        y-=0.5

        return (self.x + x*self.viewWidth, self.y + y*self.viewHeight)


