import copy
import random
import individual
import time
import sys, traceback

POP_SIZE = 15
GENERATIONS = 100
MUT_RATE = 0.01
BIG_MUT_RATE = 0.1
XO_RATE = 0.3
GENENUM = individual.GENENUM

ELITISM = 0.1

class Population:
    
    #constructor
    def __init__(self, type):
        self.trace = False

        if self.trace: traceback.print_tb(sys.last_traceback)
        t = time.localtime()

        # add below code when evolving to create the evolution output
        #self.fileName = str(t[2]) + "-" + str(t[1]) + "-" + str(t[3]) + "-" + str(t[4])+ "-" + type+ ".txt"
        #outfile = open(self.fileName, 'w')

        #self.fitnessfileName = "fitness" + str(t[2]) + "-" + str(t[1]) + "-" + str(t[3]) + "-" + str(t[4]) +"-" + type + ".txt"
        #fitnessoutfile = open(self.fitnessfileName, 'w')

        #outfile.write("Gen# \t Chromosome \n Fitness: \t Flag \t PyrCt \t EnCt \t time \n")

        random.seed()
        global POP_SIZE
        self.pop = []
        for i in range(POP_SIZE):
            if self.trace: print "individual created: " + str(i)
            tempInd = individual.Individual()
            tempInd.initialize()
            self.pop.append(tempInd)
            if self.trace: print individual.GENENUM

    def setDefenders(self):
        for i in self.pop:
            i.attacker = False

    def setChromosomes(self):
        for i in self.pop:
            i.setChromosome()

    def newPopulation(self, generation):
    # create a new population for the generation using elitism, xover
    # and mutation
        global GENENUM, XO_RATE, POP_SIZE, pop, ELITISM

        outline = str(generation) + "\t"
        # sort the population by incremental fitness prior to xover
        self.pop.sort()
        # make a copy of the population
        newpop = []
        for individual in self.pop:
            newpop.append(copy.deepcopy(individual))
        
        # sorted by best to worse for elitism
        newpop.reverse()

        # find the average fitness
        sum = 0
        for i in newpop:
            sum += i.getFitness()
        average = sum / POP_SIZE
            
        # append to fitness file: max fitness, min fitness and avg
        #fitnessFile = open(self.fitnessfileName, 'a')
        #fitnessFile.write(str(generation) + "\t" + str(newpop[0].getFitness()) + "\t" + str(newpop[-1].getFitness()) + "\t" + str(average) + "\n")

        outline += str(generation) + "\t"
        outline += newpop[0].toString()
        outline += "\n"
        outline += str(newpop[0].getFitness()) + "\t"
        outline += newpop[0].fitnessToString()
        outline += "\n"
 
        # perform xover on all non elite individuals
        for index in range((int(POP_SIZE*ELITISM)), POP_SIZE):
            if random.random() < XO_RATE:
                newpop[index] = self.xover()
        
        for index in range((int(POP_SIZE*ELITISM)), POP_SIZE):
            self.mutate(newpop)
            
        
        # replace pop with the newly formed xover pop
        for i in range(len(newpop)):
            self.pop[i] = newpop[i]

        if self.trace: print outline
        outline += "\n"
        #outfile = open(self.fileName, 'a')
        #outfile.write(outline)


    def xover(self):
    # perform two point crossover on the genes
    # NOTE: Must be sorted first 
        # retrieve the two (unique) parents
        parentindex = self.rankSelect()
        while parentindex[0] == parentindex[1]:
            parentindex = self.rankSelect()
        parent1 = self.pop[parentindex[0]]
        parent2 = self.pop[parentindex[1]]
        child = individual.Individual()
        child.initialize()
        # get the two unique crossover point
        point1 = random.randint(0,GENENUM-1)
        point2 = random.randint(0,GENENUM-1)
        while point1 == point2:
            point1 = random.randint(0,GENENUM-1)
            point2 = random.randint(0,GENENUM-1)
        if point1 > point2:
            temp = point1
            point1 = point2
            point2 = temp
        # perform xover
        for i in range(GENENUM):
            if i <= point1 or i > point2:
                child.setGene(i, parent1.getGene(i))
            else:
                child.setGene(i, parent2.getGene(i))
        return child
                                
        
    def rankSelect(self):
    # Perform ranked selection 
        global POP_SIZE
        parent = []
        # n(n+1)/2
        total = (POP_SIZE * (POP_SIZE+1)) / 2
        for i in range(2):
            randVal = random.randint(1, total)
            rankSum = 0
            for rank in range(POP_SIZE):
                rankSum += rank+1
                if randVal <= rankSum:
                    parent.append(rank)
                    break
        return parent

        
    def mutate(self, population):
    # check if each gene in each chromosome in the popuation is to be
    #  mutated. If so, there is a small chance of a big mutation
        global MUT_RATE, BIG_MUT_RATE, GENENUM, POP_SIZE, ELITISM
        elitecount = 0
        for indiv in population:
            #dont mutate the elite individuals
            if elitecount < (POP_SIZE * ELITISM):
                continue
            for index in range(GENENUM):
                if random.random() < MUT_RATE:
                # if the gene is to be mutated
                    value = indiv.getGene(index)
                    if random.random() < BIG_MUT_RATE:
                    # if the gene is to have a big mutation
                        indiv.setGene(index, random.randint(-50,50))
                    else:
                    # if the gene is to have a small mutation
                        if random.randint(0,1) == 0:
                            setValue = value + 10
                            if setValue > 50:
                                setValue -= 100
                        else:
                            setValue = value - 10
                            if setValue < -50:
                                setValue += 100
                        indiv.setGene(index, setValue)


    def getIndividual(self, index):
    # retrieve the individual at given index
        return self.pop[index]


    def setFitness(self, index, win, playerCount, enemyCount, time, oppWinRatio):
    # set the fitness for the individual at given index
        self.pop[index].setFitness(win, playerCount, enemyCount, time, oppWinRatio, globals.COEVOLVE)


    def printPopulation(self):
        for indiv in self.pop:
           print indiv.toString()

'''
if __name__=="__main__":
    testpop = Population()
    print "Before:"
    testpop.printPopulation()
    testpop.newPopulation()
    print "After:"
    testpop.printPopulation()
'''

import globals
