'''
BoatViz User Interface for The Bandit Naval warfare RTS game
'''

import globals
import vector
import graphics

#import time

class SimpleGameEngine:

    projectileCount = 0
    
    def __init__(self):
        self.boats = []
        self.projectiles = []
        self.events = []
        self.firingtimer = 0

        self.trace = False

    def reset(self):
        self.boats = []
        self.projectiles = []
        self.events = []
        self.firingtimer = 0

    
    def getPosition(self, name):
        b = self.findObject(name)
        if not b == None:
            return (b.x, b.y)
        print name    
        return None

    def getOrientation(self, name):
        b = self.findObject(name)
        if not b == None:
            return b.orientation
        return None    

    def getLength(self, name):
        return self.findBoat(name).length

    def getWidth(self, name):
        return self.findBoat(name).width

    def getHealth(self, name):
        return self.findBoat(name).health*1.0/self.findBoat(name).maxHealth

    def getDestinations(self, name):
        b = self.findBoat(name)
        if b.waypoints == []: return None
        else: return b.waypoints

    def getTargetPosition(self, name):
        b = self.findBoat(name).target
        if not b: return None
        else: return (b.x, b.y)

    def hit(self, name):
        b = self.findBoat(name)
        b.health -= 15
        if b.health < 0: 
            b.health = 0

    def setDestination(self, name, x, y, z):
        self.findBoat(name).waypoints = [(x, y)]
        if self.trace: print "Setting: " + str(x) + ", " + str(y)

    def addDestination(self, name, x, y, z):
        self.findBoat(name).waypoints.append((x, y))
        if self.trace: print "Adding: " + str(x) + ", " + str(y)

    def attack(self, name, target):
        self.findBoat(name).attack(self.findBoat(target))

    def step(self, entityManager):
        for i in self.boats:
            i.update()
        for i in self.projectiles:
            i.update()

    def makeCigarette(self, name, x, y, z):
        pass

    def makeAircraftCarrier(self, name, x, y, z):
        n = Boat("data/aircraftcarrier.dat", name, x, y)
        self.boats.append(n)
        return n

    def makeDestroyer(self, name, x, y, z):
        n = Boat("data/destroyer.dat", name, x, y)
        self.boats.append(n)
        return n

    def makeCargoship(self, name, x, y, z):
        n = Boat("data/cargoship.dat", name, x, y)
        self.boats.append(n)
        return n

    def addProjectile(self, x, y, width, length, velocity, vector, damage, target, range):
        newname = "projectile" + `self.projectileCount`
        self.projectileCount += 1

        self.projectiles.append(Projectile(newname, x, y, width, length, velocity, vector, damage, target, range))

        self.addEvent(("ADD_PROJECTILE", newname, width, length))

    def findBoat(self, name):
        for i in self.boats:
            if i.name == name:
                return i

    def findObject(self, name):
        for i in self.boats:
            if i.name == name:
                return i
        for i in self.projectiles:
            if i.name == name:
                return i
        return None        

    def addEvent(self, event):
        self.events.insert(0, event)

    def getEvent(self):
        if len(self.events) == 0:
            return None
        else:
            return self.events.pop() 
        

class Boat:

    def __init__(self, filename, name, x, y):
        f = open(filename, "r")
        data = f.readlines()

        self.imagefilename = data[3].strip() #need to strip the trailing /n off
        self.length = int(data[5].strip())
        self.width = int(data[7].strip())

        self.mass = float(data[9].strip())
        self.inertia = float(data[11].strip())
        self.friction = float(data[13].strip())
        self.rotFriction = float(data[15].strip())
        self.enginePower = float(data[17].strip())
        self.rudderEffectiveness = float(data[19].strip())
        self.keelForce = float(data[21].strip())
        self.turningCircleRad = float(data[23].strip())
        self.slowdownDist = float(data[25].strip())


        self.throttlePosition = 0
        self.rudderPosition = 0
        self.velocity = 0
        self.rotVelocity = 0


        self.maxHealth = int(data[27].strip())
        self.weaponDamage = int(data[29].strip())
        self.weaponRate = float(data[31].strip())
        self.weaponRange = float(data[33].strip())

        #Start the boat with full-health
        self.health = self.maxHealth

        self.x = x
        self.y = y

        self.orientation = 0.0

        self.waypoints = []

        self.name = name

        self.target = None

        self.weaponFiringTimer = -1

    def update(self):
        self.weaponFiringTimer += globals.dt

        #Check if its time to fire another shot
        #if time.clock() - self.weaponFiringTimer > self.weaponRate/globals.timeAccel and \
        if self.weaponFiringTimer > self.weaponRate and \
            self.target != None and self.target.health > 0:

            #Check that the target is in range
            if vector.Vector(self.x-self.target.x, self.y-self.target.y).length() < self.weaponRange:
            
                #self.weaponFiringTimer = time.clock()
                self.weaponFiringTimer = globals.dt
                globals.gameEngine.addProjectile(self.x, self.y, 10, 50, 800,
                                                vector.Vector(self.target.x - self.x, self.target.y - self.y), 
                                                self.weaponDamage, self.target, self.weaponRange)
                globals.gameEngine.addEvent(("WEAPON_ANIMATION", self.name))
            

        if not self.waypoints == []:
            #Work out what the boat should do in order to reach its destination using only the
            #throttle and rudder controls


            #v = vector.Vector(self.waypoints[0][0]-self.x, self.waypoints[0][1]-self.y)

            #If the boat is within 100 meters of any future waypoint, move to the next waypoint and
            #navigate there. If we are out of waypoints - then just stop
            wpNum = 0
            wflag = False
            for wp in self.waypoints:
                v = vector.Vector(wp[0]-self.x, wp[1]-self.y)
                if v.length() < 150:
                    wflag = True
                    break
                wpNum += 1

            if wflag:
                for i in range(wpNum+1):
                    del self.waypoints[0]

           # if v.length() < 100: 
                #del self.waypoints[0]

                if self.waypoints == []:
                    self.throttlePosition = 0
                    self.rudderPosition = 0
                    self.updatePosition()
                    return
                #else:
            v = vector.Vector(self.waypoints[0][0]-self.x, self.waypoints[0][1]-self.y)


            relativeAngle = v.angle() - self.orientation 

            #ensure the relative angle is in the -180 <-> 180 range
            relativeAngle = (relativeAngle % 360) 
            if relativeAngle > 180: relativeAngle -= 360

            #Set throttle to full speed to get to the destination as fast as possible, we will slow
            #down when required
            self.throttlePosition = 1.0 

            #Determine if to turn left or right. I leave a 10 degree empty region so the boat isn't
            #constnatly turning left/right
            if relativeAngle < -5: 
                self.rudderPosition = -1
            elif relativeAngle > 5: 
                self.rudderPosition = 1
            else: 
                self.rudderPosition = 0

            #Check if we are within the slowdown distance, and scale back the throttle so we dont
            #wildly overshoot it - TODO This doesn't work very well - there are probably better
            #ways of handling this

            #if v.length() < self.slowdownDist:
                #self.throttlePosition = v.length() / self.slowdownDist


            #Check if the destination lies within our possible turning cirlce - and if
            #so adjust the throttle position to help us reach the point
            #This method seems to be working fairly well. The way it operates is by assuming there
            #are 2 big turning circles on either side of the boat, representing regions within which
            #it is not possible for the boat to reach at full speed. This code calculated the
            #distance from the waypoint to the center of each of these turning circles (the center
            #of the turning circle is 90degrees to the left and right of the boat, at a distance of
            #turningCircleRad. If the waypoint is within the turning circles, we scale back the
            #throttle accordingly
            
            rightVec = vector.Vector(angle=(self.orientation+90))
            rightVec.multiply(self.turningCircleRad)
            rightVec.add(vector.Vector(self.x, self.y))
            rightVec.subtract(vector.Vector(self.waypoints[0][0], self.waypoints[0][1]))
            rightLen = rightVec.length()

            leftVec = vector.Vector(angle=(self.orientation-90))
            leftVec.multiply(self.turningCircleRad)
            leftVec.add(vector.Vector(self.x, self.y))
            leftVec.subtract(vector.Vector(self.waypoints[0][0], self.waypoints[0][1]))
            leftLen = leftVec.length()


            #Scale back the throttle depending on how far within the turning
            #circle the waypoint is. Right on the edge (ie rightLen ~= rad, then 
            #only needs to be slightly decreased as we can almost make the turn flat out
            if min(rightLen, leftLen) < self.turningCircleRad:
                self.throttlePosition = min(rightLen, leftLen) / self.turningCircleRad
            
            

            
            
        self.updatePosition()

    def updatePosition(self):

        # Obtain the base acceleration based on engine power 
        boatAccel = self.throttlePosition * self.enginePower;
        
        # Modify the base accel by the mass and also determine current 
           #water friction which changes with speed 
        boatAccel = boatAccel/self.mass - (self.velocity * self.friction);

        # Increase/decrease velocity appropriately 
        self.velocity += boatAccel * globals.dt;

        # Calculate rotational forces 

        #The inital rotational forces are based on engine power and rudder position (scaled by
        #appripriate values
        rotAccel = (self.throttlePosition * self.enginePower * self.rudderPosition * self.rudderEffectiveness) 

        #This applies a crude force to the rotation - boats that are goign faster can't turn
        #quite as well 
        rotAccel *= (1 - (self.keelForce * self.velocity))

        #Apply rotational friction - this ensures we don't continue to get faster forever.
        #Its important we put this after the inertia calculation otherwise the frictions limits the 
        #turning ability before the inertia takes place
        rotAccel -= (self.rotVelocity * self.rotFriction)

        #Apply inertia forces. This basically just reduces the amount of power by a fixed value so
        #it takes longer to accelerate up the rotational velocity. Ie 0.1 == 10x longer to get to
        #full turning speed. We only do this when the boat is trying to turn 'harder' - 
        #which occurs only when accel and vel are both -ve or both +ve
        if rotAccel * self.rotVelocity > 0:
            rotAccel *= self.inertia

        # Alter the rotational velocity by the calculated acceleration 
        self.rotVelocity += rotAccel * globals.dt;

        # Finally apply the rotational velocity to the current orientation 
        self.orientation += self.rotVelocity * globals.dt;

        # Keep orientation within the proper bounds
        self.orientation %= 360
        
        # Update the orientation of the boatVector to match the boatOrientation 
        # VRotate2D seems to rotate the opposite direction to our cartesian space 
        # TODO: Do some tests to make sure these stay in sync properly, if the euler
           #angle and this vector slowly get out of sync due to rounding errors and 
           #minor errors in the trig functions then things will begin to go wrong.
           #Perhaps its better to completely rebuilt the boatVector from the 
           #orientation value every turn: VRotate2D(boatOri, Vector(0, 1, 0));  

        boatVector = vector.Vector(angle = self.orientation)
        boatVector.normalize();

        #Scale the vector by the velocity so it moves the correct distance
        boatVector.multiply(self.velocity*globals.dt)

        #And add this new movement to the boats position
        self.x += boatVector.x
        self.y += boatVector.y


    def attack(self, target):
        if target != self.target:
            self.target = target
            #self.weaponFiringTimer = time.clock()
            self.weaponFiringTimer = 0


            




class Projectile:
    def __init__(self, name, x, y, width, length, velocity, vector, damage, target, range):
        self.name = name
        self.x = x
        self.y = y
        self.originx = x
        self.originy = y
        self.width = width
        self.length = length
        self.velocity = velocity
        self.vector = vector
        self.orientation = vector.angle()
        self.damage = damage
        self.target = target
        self.range = range

        self.vector.normalize()

    def update(self):
        self.vector = vector.Vector(self.target.x-self.x, self.target.y-self.y)
        self.vector.normalize()

        self.x += self.vector.x * globals.dt * self.velocity
        self.y += self.vector.y * globals.dt * self.velocity

        tempVector = vector.Vector(self.x-self.originx, self.y-self.originy)
        if tempVector.length() > self.range:
            globals.gameEngine.addEvent(("REMOVE_PROJECTILE", self.name))
            globals.gameEngine.projectiles.remove(self)
        else:    

            collision = vector.Vector(self.target.x-self.x, self.target.y-self.y)

            if collision.length() < self.target.width*3:
                globals.gameEngine.addEvent(("REMOVE_PROJECTILE", self.name))
                globals.gameEngine.addEvent(("EXPLODE_BOAT", self.target.name))
                globals.gameEngine.projectiles.remove(self)
                self.target.health -= self.damage
                if(self.target.health <= 0):
                    globals.gameEngine.addEvent(("DESTROY_BOAT", self.target.name))
                    self.target.target = None

        



