'''
BoatViz User Interface for The Bandit Naval warfare RTS game
'''

import math

class Vector:
    
    def __init__(self, x=0, y=0, angle=None):
        if angle != None:
            '''Create a vector from a polar angle - assuming positive is clockwise and
            0 is alogn the y axis '''
            self.x = math.sin(math.radians(angle))
            self.y = math.cos(math.radians(angle))
        else:
            self.x = x
            self.y = y

    def multipliedBy(self, value):
        return Vector(self.x*value, self.y*value)

    def multiply(self, value):
        self.x *= value
        self.y *= value

    def add(self, v):
        self.x += v.x
        self.y += v.y

    def subtract(self, v):
        self.x -= v.x
        self.y -= v.y

    def normalize(self):
        l = self.length()
        self.x *= 1.0 / l
        self.y *= 1.0 / l

    def length(self):
        return math.hypot(self.x, self.y)
        
    def angle(self):
        '''Converts the vector into an angle - assuming positive is clockwise'''
        return math.degrees(math.atan2(self.x, self.y))


